@@ -1,5 +1,5 @@
-#ifdef BORG_USE_LIBLZ4
-#include <lz4.h>
-#else
+#ifdef BORG_USE_BUNDLED_LZ4
 #include "lz4/lib/lz4.h"
+#else
+#include <lz4.h>
 #endif
