@@ -9,6 +9,7 @@ from enum import Enum
 
 from . import shellpattern
 from .helpers import clean_lines
+from .helpers.errors import Error
 
 
 def parse_patternfile_line(line, roots, ie_commands, fallback):
@@ -53,8 +54,11 @@ class ArgparsePatternFileAction(argparse.Action):
         Lines empty or starting with '#' after stripping whitespace on both line ends are ignored.
         """
         filename = values[0]
-        with open(filename) as f:
-            self.parse(f, args)
+        try:
+            with open(filename) as f:
+                self.parse(f, args)
+        except FileNotFoundError as e:
+            raise Error(str(e))
 
     def parse(self, fobj, args):
         load_pattern_file(fobj, args.paths, args.patterns)
@@ -143,7 +147,7 @@ class PatternMatcher:
         in self.fallback is returned (defaults to None).
 
         """
-        path = normalize_path(path)
+        path = normalize_path(path).lstrip(os.path.sep)
         # do a fast lookup for full path matches (note: we do not count such matches):
         non_existent = object()
         value = self._path_full_patterns.get(path, non_existent)
@@ -215,7 +219,7 @@ class PathFullPattern(PatternBase):
     PREFIX = "pf"
 
     def _prepare(self, pattern):
-        self.pattern = os.path.normpath(pattern)
+        self.pattern = os.path.normpath(pattern).lstrip(os.path.sep)  # sep at beginning is removed
 
     def _match(self, path):
         return path == self.pattern
@@ -235,7 +239,9 @@ class PathPrefixPattern(PatternBase):
     PREFIX = "pp"
 
     def _prepare(self, pattern):
-        self.pattern = os.path.normpath(pattern).rstrip(os.path.sep) + os.path.sep
+        sep = os.path.sep
+
+        self.pattern = (os.path.normpath(pattern).rstrip(sep) + sep).lstrip(sep)  # sep at beginning is removed
 
     def _match(self, path):
         return (path + os.path.sep).startswith(self.pattern)
@@ -253,7 +259,7 @@ class FnmatchPattern(PatternBase):
         else:
             pattern = os.path.normpath(pattern) + os.path.sep + '*'
 
-        self.pattern = pattern
+        self.pattern = pattern.lstrip(os.path.sep)  # sep at beginning is removed
 
         # fnmatch and re.match both cache compiled regular expressions.
         # Nevertheless, this is about 10 times faster.
@@ -277,7 +283,7 @@ class ShellPattern(PatternBase):
         else:
             pattern = os.path.normpath(pattern) + sep + "**" + sep + "*"
 
-        self.pattern = pattern
+        self.pattern = pattern.lstrip(sep)  # sep at beginning is removed
         self.regex = re.compile(shellpattern.translate(self.pattern))
 
     def _match(self, path):
@@ -290,7 +296,7 @@ class RegexPattern(PatternBase):
     PREFIX = "re"
 
     def _prepare(self, pattern):
-        self.pattern = pattern
+        self.pattern = pattern  # sep at beginning is NOT removed
         self.regex = re.compile(pattern)
 
     def _match(self, path):
@@ -372,7 +378,7 @@ def parse_inclexcl_command(cmd_line_str, fallback=ShellPattern):
 
     cmd = cmd_prefix_map.get(cmd_line_str[0])
     if cmd is None:
-        raise argparse.ArgumentTypeError("A pattern/command must start with any one of: %s" %
+        raise argparse.ArgumentTypeError("A pattern/command must start with anyone of: %s" %
                                          ', '.join(cmd_prefix_map))
 
     # remaining text on command-line following the command character
